# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

_Q='content'
_P='certificates'
_O='certificate'
_N='sztpd'
_M='2021-01-14'
_L='import'
_K='contentType'
_J='encoding = '
_I='utf-8'
_H='xml'
_G='json'
_F='implement'
_E='conformance-type'
_D='namespace'
_C='revision'
_B='name'
_A=None
import re,sys,pem,json,base64,yangson,datetime,textwrap,traceback,pkg_resources,xml.etree.ElementTree as ET
from enum import Enum
from aiohttp import web
from xml.dom import minidom
from pyasn1.type import tag
from pyasn1.type import univ
from urllib.parse import unquote
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5280
from pyasn1.codec.der import decoder as der_decoder
from pyasn1.codec.der import encoder as der_encoder
class RedundantQueryParameters(Exception):0
class MalformedDataPath(Exception):0
yl4errors={'ietf-yang-library:modules-state':{'module-set-id':'TBD','module':[{_B:'ietf-yang-types',_C:'2013-07-15',_D:'urn:ietf:params:xml:ns:yang:ietf-yang-types',_E:_L},{_B:'ietf-restconf',_C:'2017-01-26',_D:'urn:ietf:params:xml:ns:yang:ietf-restconf',_E:_F},{_B:'ietf-netconf-acm',_C:'2018-02-14',_D:'urn:ietf:params:xml:ns:yang:ietf-netconf-acm',_E:_L},{_B:'ietf-sztp-bootstrap-server',_C:'2019-04-30',_D:'urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server',_E:_F},{_B:'ietf-yang-structure-ext',_C:'2020-06-22',_D:'urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext',_E:_F},{_B:'ietf-sztp-csr',_C:_M,_D:'urn:ietf:params:xml:ns:yang:ietf-sztp-csr',_E:_F},{_B:'ietf-crypto-types',_C:_M,_D:'urn:ietf:params:xml:ns:yang:ietf-crypto-types',_E:_F}]}}
path=pkg_resources.resource_filename(_N,'yang')
path4errors=pkg_resources.resource_filename(_N,'yang4errors')
dm4errors=yangson.DataModel(json.dumps(yl4errors),[path4errors,path])
def gen_rc_errors(error_type,error_tag,error_app_tag=_A,error_path=_A,error_message=_A,error_info=_A):
	E=error_info;D=error_message;C=error_path;B=error_app_tag;A={};A['error-type']=error_type;A['error-tag']=error_tag
	if B is not _A:A['error-app-tag']=B
	if C is not _A:A['error-path']=C
	if D is not _A:A['error-message']=D
	if E is not _A:A['error-info']=E
	return{'ietf-restconf:errors':{'error':[A]}}
def enc_rc_errors(encoding,errors_obj):
	B=errors_obj;A=encoding
	if A==_G:return json.dumps(B,indent=2)
	if A==_H:C=dm4errors.from_raw(B);D=C.to_xml();E=ET.tostring(D).decode(_I);F=minidom.parseString(E);return F.toprettyxml(indent='  ')
	raise NotImplementedError(_J+A)
class Encoding(Enum):json=1;xml=2
def obj_to_encoded_str(obj,enc,dm,sn):
	A=enc
	if A==Encoding.json:return json.dumps(obj,indent=2)
	if A==Encoding.xml:B=sn.from_raw(obj);C=yangson.instance.RootNode(B,sn,dm.schema_data,B.timestamp);D=C.to_xml();E=ET.tostring(D).decode(_I);F=minidom.parseString(E);return F.toprettyxml(indent='  ')
	raise NotImplementedError(_J+A)
def encoded_str_to_obj(estr,enc,dm,sn):
	D="Doesn't match schema: ";C=True
	if enc==Encoding.json:
		try:E=json.loads(estr)
		except Exception as A:raise Exception('JSON malformed: '+str(A))
		try:B=sn.from_raw(E,allow_nodata=C)
		except Exception as A:raise Exception(D+str(A))
	elif enc==Encoding.xml:
		try:F=ET.fromstring(estr)
		except Exception as A:raise Exception('XML malformed: '+str(A))
		try:B=sn.from_xml(F,isroot=C,allow_nodata=C)
		except Exception as A:raise Exception(D+str(A))
	else:raise NotImplementedError(_J+encoding)
	try:G=yangson.instance.RootNode(B,sn,dm.schema_data,B.timestamp)
	except Exception as A:raise Exception(D+str(A))
	try:H=G.raw_value()
	except Exception as A:raise Exception('Error transcoding: '+str(A))
	return H
def multipart_pem_to_der_dict(multipart_pem):
	A={};E=pem.parse(bytes(multipart_pem,_I))
	for F in E:
		C=F.as_text().splitlines();D=base64.b64decode(''.join(C[1:-1]));B=re.sub('-----BEGIN (.*)-----','\\g<1>',C[0])
		if B not in A:A[B]=[D]
		else:A[B].append(D)
	return A
def der_dict_to_multipart_pem(der_dict):
	H='-----\n';C=der_dict;A='';D=C.keys()
	for B in D:
		E=C[B]
		for F in E:G=base64.b64encode(F).decode('ASCII');A+='-----BEGIN '+B+H;A+=textwrap.fill(G,64)+'\n';A+='-----END '+B+H
	return A
def ders_to_degenerate_cms_obj(cert_ders):
	B=rfc5652.CertificateSet().subtype(implicitTag=tag.Tag(tag.tagClassContext,tag.tagFormatSimple,0))
	for E in cert_ders:F,G=der_decoder.decode(E,asn1Spec=rfc5280.Certificate());assert not G;D=rfc5652.CertificateChoices();D[_O]=F;B[len(B)]=D
	A=rfc5652.SignedData();A['version']=1;A['digestAlgorithms']=rfc5652.DigestAlgorithmIdentifiers().clear();A['encapContentInfo']['eContentType']=rfc5652.id_data;A[_P]=B;C=rfc5652.ContentInfo();C[_K]=rfc5652.id_signedData;C[_Q]=der_encoder.encode(A);return C
def degenerate_cms_obj_to_ders(cms_obj):
	A=cms_obj
	if A[_K]!=rfc5652.id_signedData:raise KeyError('unexpected content type: '+str(A[_K]))
	D,H=der_decoder.decode(A[_Q],asn1Spec=rfc5652.SignedData());E=D[_P];B=[]
	for F in E:C=F[_O];assert type(C)==rfc5280.Certificate;G=der_encoder.encode(C);B.append(G)
	return B
def parse_raw_path(full_raw_path):
	P="' appears more than once.  RFC 8040, Section 4.8 states that each parameter can appear at most once.";O="Query parameter '";L='?';G=full_raw_path;F='=';B='/'
	if L in G:assert G.count(L)==1;A,J=G.split(L)
	else:A=G;J=_A
	if A=='':A=B
	elif A[0]!=B:raise MalformedDataPath("The datastore-specific part of the path, when present, must begin with a '/' character.")
	elif A[-1]==B:raise MalformedDataPath("Trailing '/' characters are not supported.")
	if A==B:H=B
	else:
		H='';M=A[1:].split(B)
		for D in M:
			if D=='':raise MalformedDataPath("The data path contains a superflous '/' character.")
			if F in D:assert D.count(F)==1;C,K=D.split(F);H+=B+unquote(C)+F+K
			else:H+=B+unquote(D)
	E=dict()
	if J is not _A:
		N=J.split('&')
		for I in N:
			if F in I:
				C,K=I.split(F,1)
				if C in E:raise RedundantQueryParameters(O+C+P)
				E[unquote(C)]=K
			else:
				if I in E:raise RedundantQueryParameters(O+C+P)
				E[unquote(I)]=_A
	return H,E
def check_http_headers(request,supported_media_types,accept_required):
	Q='application/yang-data+xml';P='The request method (';O='missing-attribute';N=accept_required;M='".';L='" or "';K='protocol';J='application/yang-data+json';H='Content-Type';G='text/plain';F=supported_media_types;D='Accept';B=request;assert type(F)==tuple
	if not B.body_exists:
		if any((B.method==A for A in('PUT','PATCH')))or B.method=='POST'and'ietf-datastores:running'in B.path:
			A=web.Response(status=400);C=P+B.method+') must include a request body.'
			if not D in B.headers or not any((B.headers[D]==A for A in F)):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,O,error_message=C)
				if A.content_type==J:A.text=enc_rc_errors(_G,E)
				else:A.text=enc_rc_errors(_H,E)
				return A,E
	if B.body_exists:
		if any((B.method==A for A in('GET','DELETE'))):
			A=web.Response(status=400);C=P+B.method+') should not include a request body. (enforced here)'
			if not D in B.headers or not any((B.headers[D]==A for A in F)):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,O,error_message=C)
				if A.content_type==J:A.text=enc_rc_errors(_G,E)
				else:A.text=enc_rc_errors(_H,E)
				return A,E
		if not H in B.headers:
			A=web.Response(status=400);C='A "Content-Type" value must be specified when a request body is passed. The "Content-Type" value must be "application/yang-data+json" or "application/yang-data+xml".'
			if not D in B.headers or not any((B.headers[D]==A for A in F)):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,O,error_message=C)
				if A.content_type==J:A.text=enc_rc_errors(_G,E)
				else:A.text=enc_rc_errors(_H,E)
				return A,E
		if not any((B.headers[H]==A for A in F)):
			A=web.Response(status=400);C='The "Content-Type" value, when specified, must be "'+L.join(F)+'". Got: "'+B.headers[H]+M
			if not D in B.headers or not any((B.headers[D]==A for A in(J,Q))):A.content_type=G;A.text=C;return A,C
			else:
				A.content_type=B.headers[D];E=gen_rc_errors(K,'bad-attribute',error_message=C)
				if A.content_type==J:A.text=enc_rc_errors(_G,E)
				else:A.text=enc_rc_errors(_H,E)
				return A,E
	I=_A
	if D not in B.headers or B.headers[D]=='*/*':
		if H not in B.headers:
			if N:A=web.Response(status=406);C='Unable to determine response encoding; neither "Accept" nor "Content-Type" specified.  An "Accept" value should be specified, and have the value "'+L.join(F)+M;A.content_type=G;A.text=C;return A,C
			I=G
		elif not any((B.headers[H]==A for A in(J,Q))):
			if N:A=web.Response(status=406);C='Unable to determine response encoding; "Accept" not specified and the "Content-Type" specified ('+B.headers[H]+') is invalid.  An "Accept" value should be specified, and have the value "'+L.join(F)+M;A.content_type=G;A.text=C;return A,C
			I=G
		else:I=B.headers[H]
	elif not any((B.headers[D]==A for A in F)):A=web.Response(status=406);C='Unable to determine response encoding; the "Accept" value specified ('+B.headers[D]+') is invalid.  The "Accept" value, when specified, must have the value "'+L.join(F)+M;A.content_type=G;A.text=C;return A,C
	else:I=B.headers[D]
	return I,_A


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_Z='wn-sztpd-x:'
_Y='Unable to parse "input" JSON document: '
_X='malformed-message'
_W='wn-sztpd-x'
_V='application/yang-data+json'
_U='/wn-sztpd-x:tenants/tenant=[^ ]*'
_T='/wn-sztpd-x:tenants/tenant=[^/]*/'
_S='Top node names must begin with the "wn-sztpd-1" prefix.'
_R='application'
_Q='wn-sztpd-1'
_P=False
_O='name'
_N='wn-sztpd-1:'
_M=True
_L='/wn-sztpd-x:tenants/tenant/0/'
_K='Non-root data_paths must begin with "/wn-sztpd-1:".'
_J='wn-sztpd-x:tenant'
_I='+'
_H='invalid-value'
_G='text/plain'
_F='protocol'
_E=None
_D=':'
_C='/wn-sztpd-x:tenants/tenant='
_B='/wn-sztpd-1:'
_A='/'
import re,json,datetime,basicauth
from aiohttp import web
from passlib.hash import sha256_crypt
from .  import yl
from .  import dal
from .  import utils
from .rcsvr import RestconfServer
from .handler import RouteHandler
class TenantViewHandler(RouteHandler):
	supported_media_types=_V,
	def __init__(A,native):A.native=native
	async def _check_auth(C,request,data_path):
		K='access-denied';J='comment';I='failure';H='outcome';E=request;A={};A['timestamp']=datetime.datetime.utcnow();A['source-ip']=E.remote;A['source-proxies']=list(E.forwarded);A['host']=E.host;A['method']=E.method;A['path']=E.path;L=E.headers.get('AUTHORIZATION')
		if L is _E:A[H]=I;A[J]='No authorization specified in the HTTP header.';await C.native._insert_audit_log_entry(_E,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D,indent=2);return B
		G,N=basicauth.decode(L);F=_E
		try:F=await C.native.dal.get_tenant_name_for_admin(G)
		except dal.NodeNotFound as Q:A[H]=I;A[J]='Unknown admin: '+G;await C.native._insert_audit_log_entry(_E,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D,indent=2);return B
		if F==_E:A[H]=I;A[J]='Host-level admins cannot use tenant interface ('+G+').';await C.native._insert_audit_log_entry(_E,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D,indent=2);return B
		O=_A+C.native.dal.app_ns+':tenants/tenant='+F+'/admin-accounts/admin-account='+G+'/password';P=await C.native.dal.handle_get_config_request(O);M=P[C.native.dal.app_ns+':password'];assert M.startswith('$5$')
		if not sha256_crypt.verify(N,M):A[H]=I;A[J]='Password mismatch for admin '+G;await C.native._insert_audit_log_entry(F,A);B=web.Response(status=401);D=utils.gen_rc_errors(_F,K);B.text=json.dumps(D,indent=2);return B
		A[H]='success';await C.native._insert_audit_log_entry(F,A);return F
	async def handle_get_restconf_root(E,request):
		F=request;H=_A;C=await E._check_auth(F,H)
		if type(C)is web.Response:A=C;return A
		else:I=C
		B,J=utils.check_http_headers(F,E.supported_media_types,accept_required=_M)
		if type(B)is web.Response:A=B;return A
		else:assert type(B)==str;D=B;assert D!=_G;G=utils.Encoding[D.rsplit(_I,1)[1]]
		A=web.Response(status=200);A.content_type=D
		if G==utils.Encoding.json:A.text='{\n    "ietf-restconf:restconf" : {\n        "data" : {},\n        "operations" : {},\n        "yang-library-version" : "2019-01-04"\n    }\n}\n'
		else:assert G==utils.Encoding.xml;A.text='<restconf xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">\n    <data/>\n    <operations/>\n    <yang-library-version>2016-06-21</yang-library-version>\n</restconf>\n'
		return A
	async def handle_get_yang_library_version(E,request):
		F=request;H=_A;C=await E._check_auth(F,H)
		if type(C)is web.Response:A=C;return A
		else:I=C
		B,J=utils.check_http_headers(F,E.supported_media_types,accept_required=_M)
		if type(B)is web.Response:A=B;return A
		else:assert type(B)==str;D=B;assert D!=_G;G=utils.Encoding[D.rsplit(_I,1)[1]]
		A=web.Response(status=200);A.content_type=D
		if G==utils.Encoding.json:A.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}'
		else:assert G==utils.Encoding.xml;A.text='<yang-library-version xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">2019-01-04</yang-library-version>'
		return A
	async def handle_get_opstate_request(E,request):
		F=request;B,M=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_operational:]);G=await E._check_auth(F,B)
		if type(G)is web.Response:A=G;return A
		else:H=G
		D,V=utils.check_http_headers(F,E.supported_media_types,accept_required=_M)
		if type(D)is web.Response:A=D;return A
		else:assert type(D)==str;N=D;assert N!=_G;W=utils.Encoding[N.rsplit(_I,1)[1]]
		if B=='/ietf-yang-library:yang-library':A=web.Response(status=200);A.content_type=_V;A.text=getattr(yl,'nbi_x_tenant')();return A
		assert B==_A or B.startswith(_B)
		if B==_A:O=_C+H
		else:
			if not B.startswith(_B):A=web.Response(status=400);S=utils.gen_rc_errors(_F,_H,error_message=_K);A.text=json.dumps(S,indent=2);return A
			X,P=B.split(_D,1);assert P!=_E;O=_C+H+_A+P
		Q=dict()
		for R in M.keys():Q[R]=re.sub(_B,_C+H+_A,M[R])
		I,C=await E.native.handle_get_opstate_request_lower_half(O,Q)
		if C!=_E:
			assert I.status==200;J={}
			if B==_A:
				for K in C[_J][0].keys():
					if K==_O:continue
					J[_N+K]=C[_J][0][K]
			else:L=next(iter(C));assert L.count(_D)==1;T,U=L.split(_D);assert T==_W;assert type(C)==dict;assert len(C)==1;J[_N+U]=C[L]
			I.text=json.dumps(J,indent=2)
		return I
	async def handle_get_config_request(E,request):
		F=request;A,M=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_running:]);G=await E._check_auth(F,A)
		if type(G)is web.Response:C=G;return C
		else:H=G
		D,V=utils.check_http_headers(F,E.supported_media_types,accept_required=_M)
		if type(D)is web.Response:C=D;return C
		else:assert type(D)==str;N=D;assert N!=_G;W=utils.Encoding[N.rsplit(_I,1)[1]]
		assert A==_A or A.startswith(_B)
		if A==_A:O=_C+H
		else:
			if not A.startswith(_B):C=web.Response(status=400);S=utils.gen_rc_errors(_F,_H,error_message=_K);C.text=json.dumps(S,indent=2);return C
			X,P=A.split(_D,1);assert P!=_E;O=_C+H+_A+P
		Q=dict()
		for R in M.keys():Q[R]=re.sub(_B,_C+H+_A,M[R])
		I,B=await E.native.handle_get_config_request_lower_half(O,Q)
		if B!=_E:
			assert I.status==200;J={}
			if A==_A:
				for K in B[_J][0].keys():
					if K==_O:continue
					J[_N+K]=B[_J][0][K]
			else:L=next(iter(B));assert L.count(_D)==1;T,U=L.split(_D);assert T==_W;assert type(B)==dict;assert len(B)==1;J[_N+U]=B[L]
			I.text=json.dumps(J,indent=2)
		return I
	async def handle_post_config_request(G,request):
		C=request;D,K=utils.parse_raw_path(C._message.path[RestconfServer.len_prefix_running:]);H=await G._check_auth(C,D)
		if type(H)is web.Response:A=H;return A
		else:I=H
		E,U=utils.check_http_headers(C,G.supported_media_types,accept_required=_P)
		if type(E)is web.Response:A=E;return A
		else:assert type(E)==str;L=E;assert L!=_G;V=utils.Encoding[L.rsplit(_I,1)[1]]
		if D==_A:M=_C+I
		else:
			if not D.startswith(_B):A=web.Response(status=400);B=utils.gen_rc_errors(_F,_H,error_message=_K);A.text=json.dumps(B,indent=2);return A
			W,N=D.split(_D,1);assert N!=_E;M=_C+I+_A+N
		O=dict()
		for P in K.keys():O[P]=re.sub(_B,_C+I+_A,K[P])
		try:F=await C.json()
		except json.decoder.JSONDecodeError as Q:A=web.Response(status=400);B=utils.gen_rc_errors(_F,_X,error_message=_Y+str(Q));A.text=json.dumps(B,indent=2);return A
		assert type(F)==dict;assert len(F)==1;J=next(iter(F));assert J.count(_D)==1;R,S=J.split(_D)
		if R!=_Q:A=web.Response(status=400);B=utils.gen_rc_errors(_R,_H,error_message=_S);A.text=json.dumps(B,indent=2);return A
		T={_Z+S:F[J]};A=await G.native.handle_post_config_request_lower_half(M,O,T)
		if A.status!=201:
			if'/wn-sztpdex:tenants/tenant/0/'in A.text:A.text=A.text.replace(_L,_B)
			elif _C in A.text:A.text=re.sub(_T,_B,A.text);A.text=re.sub(_U,_B,A.text)
		return A
	async def handle_put_config_request(H,request):
		E=request;D,M=utils.parse_raw_path(E._message.path[RestconfServer.len_prefix_running:]);I=await H._check_auth(E,D)
		if type(I)is web.Response:A=I;return A
		else:F=I
		G,X=utils.check_http_headers(E,H.supported_media_types,accept_required=_P)
		if type(G)is web.Response:A=G;return A
		else:assert type(G)==str;N=G;assert N!=_G;Y=utils.Encoding[N.rsplit(_I,1)[1]]
		if D==_A:O=_C+F
		else:
			if not D.startswith(_B):A=web.Response(status=400);B=utils.gen_rc_errors(_F,_H,error_message=_K);A.text=json.dumps(B,indent=2);return A
			Z,P=D.split(_D,1);assert P!=_E;O=_C+F+_A+P
		Q=dict()
		for R in M.keys():Q[R]=re.sub(_B,_C+F+_A,M[R])
		try:C=await E.json()
		except json.decoder.JSONDecodeError as S:A=web.Response(status=400);B=utils.gen_rc_errors(_F,_X,error_message=_Y+str(S));A.text=json.dumps(B,indent=2);return A
		if D==_A:
			J={_J:[{_O:F}]}
			for K in C.keys():
				assert K.count(_D)==1;T,U=K.split(_D)
				if T!=_Q:A=web.Response(status=400);B=utils.gen_rc_errors(_R,_H,error_message=_S);A.text=json.dumps(B,indent=2);return A
				J[_J][0][U]=C[K]
		else:
			assert type(C)==dict;assert len(C)==1;L=next(iter(C));assert L.count(_D)==1;V,W=L.split(_D)
			if V!=_Q:A=web.Response(status=400);B=utils.gen_rc_errors(_R,_H,error_message=_S);A.text=json.dumps(B,indent=2);return A
			J={_Z+W:C[L]}
		A=await H.native.handle_put_config_request_lower_half(O,Q,J)
		if A.status!=201 and A.status!=204:
			if _L in A.text:A.text=A.text.replace(_L,_B)
			elif _C in A.text:A.text=re.sub(_T,_B,A.text);A.text=re.sub(_U,_B,A.text)
		return A
	async def handle_delete_config_request(E,request):
		F=request;B,N=utils.parse_raw_path(F._message.path[RestconfServer.len_prefix_running:]);G=await E._check_auth(F,B)
		if type(G)is web.Response:A=G;return A
		else:H=G
		C,O=utils.check_http_headers(F,E.supported_media_types,accept_required=_P)
		if type(C)is web.Response:A=C;return A
		else:
			assert type(C)==str;D=C
			if D==_G:L=_E
			else:L=utils.Encoding[D.rsplit(_I,1)[1]]
		if B==_A:I=_C+H
		else:
			if not B.startswith(_B):
				J=_K;A=web.Response(status=400);A.content_type=D
				if D==_G:A.text=J
				else:M=utils.gen_rc_errors(_F,_H,error_message=J);A.text=json.dumps(M,indent=2)
				return A
			P,K=B.split(_D,1);assert K!=_E;I=_C+H+_A+K
		A=await E.native.handle_delete_config_request_lower_half(I)
		if A.status!=204:
			if _L in A.text:A.text=A.text.replace(_L,_B)
			elif _C in A.text:A.text=re.sub(_T,_B,A.text);A.text=re.sub(_U,_B,A.text)
		return A
	async def handle_action_request(A,request):0
	async def handle_rpc_request(A,request):0


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_A3=' found): '
_A2='The asymmetric-key has a mismatched public/private key pair: '
_A1=b"this is some data I'd like to sign"
_A0='Unsupported "private-key-format" ('
_z='ietf-crypto-types:ec-private-key-format'
_y='cleartext-private-key'
_x='Parsing public key structure failed for '
_w='Unsupported "public-key-format" ('
_v='ietf-crypto-types:subject-public-key-info-format'
_u='should check to see if an alarm can be cleared...'
_t='admin-account'
_s='$0$'
_r='%Y-%m-%dT%H:%M:%SZ'
_q='password-last-modified'
_p='module'
_o='sztpd.plugins.'
_n="why wasn't this assertion caught by val? "
_m='operation-failed'
_l='data-exists'
_k='missing-attribute'
_j='Unable to parse "input" JSON document: '
_i='malformed-message'
_h='method'
_g='": '
_f='cert-data'
_e='public-key'
_d='\\g<1>'
_c='.*plugins/plugin=([^/]*).*'
_b='function'
_a='application/yang-data+json'
_Z='private-key-format'
_Y=') for '
_X='public-key-format'
_W='need to implement this code'
_V='SZTPD_MODE'
_U='plugin'
_T=')'
_S=' ('
_R=False
_Q='operation-not-supported'
_P=True
_O='functions'
_N='name'
_M='certificate'
_L='certificates'
_K='+'
_J='text/plain'
_I='password'
_H='unknown-element'
_G='sleep'
_F='invalid-value'
_E='application'
_D='protocol'
_C='asymmetric-key'
_B=None
_A='/'
import os,re,sys,json,base64,signal,asyncio,yangson,datetime,basicauth,importlib,pkg_resources
from enum import Enum
from aiohttp import web
from enum import IntFlag
from fifolock import FifoLock
from passlib.hash import sha256_crypt
from .dal import DataAccessLayer
from .val import ValidationLayer
from .rcsvr import RestconfServer
from .handler import RouteHandler
from .  import dal
from .  import val
from .  import utils
from pyasn1.codec.der.encoder import encode as encode_der
from pyasn1.codec.der.decoder import decode as decode_der
from pyasn1.error import PyAsn1Error
from pyasn1_modules import rfc4055
from pyasn1_modules import rfc5280
from pyasn1_modules import rfc5480
from pyasn1_modules import rfc5915
from pyasn1_modules import rfc5652
from pyasn1.type import univ
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import InvalidSignature
from cryptography.x509.oid import ExtensionOID
class RefAction(IntFlag):ADDED=1;REMOVED=2
class TimeUnit(Enum):Days=2;Hours=1;Minutes=0
class Period:
	def __init__(A,amount,units):A.amount=amount;A.units=units
class PluginNotFound(Exception):0
class PluginSyntaxError(Exception):0
class FunctionNotFound(Exception):0
class FunctionNotCallable(Exception):0
class Read(asyncio.Future):
	@staticmethod
	def is_compatible(holds):return not holds[Write]
class Write(asyncio.Future):
	@staticmethod
	def is_compatible(holds):A=holds;return not A[Read]and not A[Write]
class NativeViewHandler(RouteHandler):
	len_prefix_running=RestconfServer.len_prefix_running;len_prefix_operational=RestconfServer.len_prefix_operational;len_prefix_operations=RestconfServer.len_prefix_operations;supported_media_types=_a,
	def __init__(A,_dal,_mode,_loop):
		W=':tenants/tenant/truststore/certificate-bags/certificate-bag/certificate/cert-data';V=':tenants/tenant/keystore/asymmetric-keys/asymmetric-key/certificates/certificate/cert-data';U=':tenants/tenant/keystore/asymmetric-keys/asymmetric-key/cleartext-private-key';T=':tenants/tenant/keystore/asymmetric-keys/asymmetric-key/public-key';S=':truststore/certificate-bags/certificate-bag/certificate/cert-data';R=':keystore/asymmetric-keys/asymmetric-key/certificates/certificate/cert-data';Q=':keystore/asymmetric-keys/asymmetric-key/cleartext-private-key';P=':keystore/asymmetric-keys/asymmetric-key/public-key';O=':preferences/system/plugins/plugin/functions/function';N=':preferences/system/plugins/plugin';M=':tenants/tenant/admin-accounts/admin-account/password';L=':admin-accounts/admin-account/password';K=':plugins';A.dal=_dal;A.mode=_mode;A.loop=_loop;A.fifolock=FifoLock();A.create_callbacks={};A.change_callbacks={};A.delete_callbacks={};A.subtree_change_callbacks={};A.somehow_change_callbacks={};A.leafref_callbacks={};A.periodic_callbacks={};A.onetime_callbacks={};A.plugins={};B=A.dal.handle_get_opstate_request('/ietf-yang-library:yang-library');F=A.loop.run_until_complete(B);G=pkg_resources.resource_filename('sztpd','yang/');A.dm=yangson.DataModel(json.dumps(F),[G]);A.val=ValidationLayer(A.dm,A.dal);B=A.dal.handle_get_opstate_request(_A+A.dal.app_ns+':preferences/system/plugins')
		try:D=A.loop.run_until_complete(B)
		except dal.NodeNotFound:pass
		else:
			if _U in D[A.dal.app_ns+K]:
				for C in D[A.dal.app_ns+K][_U]:
					H=C[_N];B=_handle_plugin_created('',{_U:C},'',A);A.loop.run_until_complete(B)
					if _O in C:
						for E in C[_O][_b]:X=E[_N];I='FOO/plugins/plugin='+H+'/BAR';B=_handle_function_created('',{_b:E},I,A);A.loop.run_until_complete(B)
		A.register_create_callback(_A+A.dal.app_ns+L,_handle_admin_passwd_created);A.register_change_callback(_A+A.dal.app_ns+L,_handle_admin_passwd_changed)
		if A.mode=='x':A.register_create_callback(_A+A.dal.app_ns+M,_handle_admin_passwd_created);A.register_change_callback(_A+A.dal.app_ns+M,_handle_admin_passwd_changed)
		A.register_create_callback(_A+A.dal.app_ns+':tenants/tenant',_handle_tenant_created);A.register_create_callback(_A+A.dal.app_ns+N,_handle_plugin_created);A.register_delete_callback(_A+A.dal.app_ns+N,_handle_plugin_deleted);A.register_create_callback(_A+A.dal.app_ns+O,_handle_function_created);A.register_delete_callback(_A+A.dal.app_ns+O,_handle_function_deleted);A.register_create_callback(_A+A.dal.app_ns+P,_handle_asymmetric_public_key_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+P,_handle_asymmetric_public_key_created_or_changed);A.register_create_callback(_A+A.dal.app_ns+Q,_handle_asymmetric_private_key_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+Q,_handle_asymmetric_private_key_created_or_changed);A.register_create_callback(_A+A.dal.app_ns+R,_handle_asymmetric_key_cert_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+R,_handle_asymmetric_key_cert_created_or_changed);A.register_create_callback(_A+A.dal.app_ns+S,_handle_trust_anchor_cert_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+S,_handle_trust_anchor_cert_created_or_changed)
		if A.mode=='x':A.register_create_callback(_A+A.dal.app_ns+T,_handle_asymmetric_public_key_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+T,_handle_asymmetric_public_key_created_or_changed);A.register_create_callback(_A+A.dal.app_ns+U,_handle_asymmetric_private_key_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+U,_handle_asymmetric_private_key_created_or_changed);A.register_create_callback(_A+A.dal.app_ns+V,_handle_asymmetric_key_cert_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+V,_handle_asymmetric_key_cert_created_or_changed);A.register_create_callback(_A+A.dal.app_ns+W,_handle_trust_anchor_cert_created_or_changed);A.register_change_callback(_A+A.dal.app_ns+W,_handle_trust_anchor_cert_created_or_changed)
		A.register_change_callback(_A+A.dal.app_ns+':transport/listen',_handle_transport_changed);A.register_delete_callback(_A+A.dal.app_ns+':transport',_handle_transport_delete);A.register_periodic_callback(Period(24,TimeUnit.Hours),datetime.datetime(2000,1,1,0),_check_expirations)
		for J in A.dal.ref_stat_collectors:A.register_create_callback(J.replace('/reference-statistics',''),_handle_ref_stat_parent_created)
	def register_create_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.create_callbacks:A.create_callbacks[B]=[C]
		else:A.create_callbacks[B].append(C)
	def register_change_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.change_callbacks:A.change_callbacks[B]=[C]
		else:A.change_callbacks[B].append(C)
	def register_subtree_change_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.subtree_change_callbacks:A.subtree_change_callbacks[B]=[C]
		else:A.subtree_change_callbacks[B].append(C)
	def register_somehow_change_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.somehow_change_callbacks:A.somehow_change_callbacks[B]=[C]
		else:A.somehow_change_callbacks[B].append(C)
	def register_delete_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.delete_callbacks:A.delete_callbacks[B]=[C]
		else:A.delete_callbacks[B].append(C)
	def register_onetime_callback(A,timestamp,callback,opaque):
		B=callback
		if schema_path not in A.onetime_callbacks:A.onetime_callbacks[schema_path]=[B]
		else:A.onetime_callbacks[schema_path].append(B)
	def register_periodic_callback(A,period,anchor,callback):0
	def register_leafref_callback(A,schema_path,callback):
		C=callback;B=schema_path
		if B not in A.leafref_callbacks:A.leafref_callbacks[B]=[C]
		else:A.leafref_callbacks[B].append(C)
	async def _insert_audit_log_entry(A,tenant_name,audit_log_entry):
		C=audit_log_entry;B=tenant_name
		if C[_h]in{'GET','HEAD'}:return
		if B==_B:D=_A+A.dal.app_ns+':audit-log'
		else:F=A.dal.opaque();assert F=='x';D=_A+A.dal.app_ns+':tenants/tenant='+B+'/audit-log'
		E={};E[A.dal.app_ns+':log-entry']=C;await A.dal.handle_post_opstate_request(D,E)
	async def _check_auth(B,request,data_path):
		S='No authorization required for fresh installs.';R=':admin-accounts/admin-account';N='access-denied';M='failure';L='success';G='comment';F='outcome';D=request;A={};A['timestamp']=datetime.datetime.utcnow();A['source-ip']=D.remote;A['source-proxies']=list(D.forwarded);A['host']=D.host;A[_h]=D.method;A['path']=D.path;J=D.headers.get('AUTHORIZATION')
		if J is _B:
			H=await B.dal.num_elements_in_list(_A+B.dal.app_ns+R)
			if H==0:A[F]=L;A[G]=S;await B._insert_audit_log_entry(_B,A);return web.Response(status=200)
			A[F]=M;A[G]='No authorization specified in the HTTP header.';await B._insert_audit_log_entry(_B,A);C=web.Response(status=401);E=utils.gen_rc_errors(_D,N);C.text=json.dumps(E,indent=2);return C
		I,O=basicauth.decode(J);P=_A+B.dal.app_ns+':admin-accounts/admin-account='+I+'/password'
		try:Q=await B.dal.handle_get_config_request(P)
		except dal.NodeNotFound as T:
			H=await B.dal.num_elements_in_list(_A+B.dal.app_ns+R)
			if H==0:A[F]=L;A[G]=S;await B._insert_audit_log_entry(_B,A);return web.Response(status=200)
			A[F]=M;A[G]='Unknown admin: '+I;await B._insert_audit_log_entry(_B,A);C=web.Response(status=401);E=utils.gen_rc_errors(_D,N);C.text=json.dumps(E,indent=2);return C
		K=Q[B.dal.app_ns+':password'];assert K.startswith('$5$')
		if not sha256_crypt.verify(O,K):A[F]=M;A[G]='Password mismatch for admin '+I;await B._insert_audit_log_entry(_B,A);C=web.Response(status=401);E=utils.gen_rc_errors(_D,N);C.text=json.dumps(E,indent=2);return C
		A[F]=L;await B._insert_audit_log_entry(_B,A);return web.Response(status=200)
	async def handle_get_restconf_root(D,request):
		E=request;G=_A;A=await D._check_auth(E,G)
		if A.status==401:return A
		B,H=utils.check_http_headers(E,D.supported_media_types,accept_required=_P)
		if type(B)is web.Response:A=B;return A
		else:assert type(B)==str;C=B;assert C!=_J;F=utils.Encoding[C.rsplit(_K,1)[1]]
		A=web.Response(status=200);A.content_type=C
		if F==utils.Encoding.json:A.text='{\n    "ietf-restconf:restconf" : {\n        "data" : {},\n        "operations" : {},\n        "yang-library-version" : "2019-01-04"\n    }\n}\n'
		else:assert F==utils.Encoding.xml;A.text='<restconf xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">\n    <data/>\n    <operations/>\n    <yang-library-version>2016-06-21</yang-library-version>\n</restconf>\n'
		return A
	async def handle_get_yang_library_version(D,request):
		E=request;G=_A;A=await D._check_auth(E,G)
		if A.status==401:return A
		B,H=utils.check_http_headers(E,D.supported_media_types,accept_required=_P)
		if type(B)is web.Response:A=B;return A
		else:assert type(B)==str;C=B;assert C!=_J;F=utils.Encoding[C.rsplit(_K,1)[1]]
		A=web.Response(status=200);A.content_type=C
		if F==utils.Encoding.json:A.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}'
		else:assert F==utils.Encoding.xml;A.text='<yang-library-version xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">2019-01-04</yang-library-version>'
		return A
	async def handle_get_opstate_request(C,request):
		D=request;E,H=utils.parse_raw_path(D._message.path[RestconfServer.len_prefix_operational:]);A=await C._check_auth(D,E)
		if A.status==401:return A
		B,J=utils.check_http_headers(D,C.supported_media_types,accept_required=_P)
		if type(B)is web.Response:I=B;return I
		else:assert type(B)==str;F=B;assert F!=_J;K=utils.Encoding[F.rsplit(_K,1)[1]]
		A,G=await C.handle_get_opstate_request_lower_half(E,H)
		if G!=_B:A.text=json.dumps(G,indent=2)
		return A
	async def handle_get_opstate_request_lower_half(D,data_path,query_dict):
		B=query_dict
		async with D.fifolock(Read):
			if os.environ.get(_V)and _G in B:await asyncio.sleep(int(B[_G]))
			try:F=await D.dal.handle_get_opstate_request(data_path,B)
			except dal.NodeNotFound as C:A=web.Response(status=404);E=utils.gen_rc_errors(_D,_H,error_message=str(C));A.text=json.dumps(E,indent=2);return A,_B
			except NotImplementedError as C:A=web.Response(status=501);E=utils.gen_rc_errors(_E,_Q,error_message=str(C));A.text=json.dumps(resp_text_ob,indent=2j);return A,_B
			A=web.Response(status=200);A.content_type=_a;return A,F
	async def handle_get_config_request(C,request):
		D=request;E,H=utils.parse_raw_path(D._message.path[RestconfServer.len_prefix_running:]);A=await C._check_auth(D,E)
		if A.status==401:return A
		B,J=utils.check_http_headers(D,C.supported_media_types,accept_required=_P)
		if type(B)is web.Response:I=B;return I
		else:assert type(B)==str;F=B;assert F!=_J;K=utils.Encoding[F.rsplit(_K,1)[1]]
		A,G=await C.handle_get_config_request_lower_half(E,H)
		if G!=_B:A.text=json.dumps(G,indent=2)
		return A
	async def handle_get_config_request_lower_half(E,data_path,query_dict):
		F=data_path;D=query_dict
		async with E.fifolock(Read):
			try:await E.val.handle_get_config_request(F,D)
			except val.InvalidDataPath as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A,_B
			except val.NonexistentSchemaNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A,_B
			except val.NodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_D,_H,error_message=str(B));A.text=json.dumps(C,indent=2);return A,_B
			if os.environ.get(_V)and _G in D:await asyncio.sleep(int(D[_G]))
			try:G=await E.dal.handle_get_config_request(F,D)
			except dal.NodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_D,_H,error_message=str(B));A.text=json.dumps(C,indent=2);return A,_B
			A=web.Response(status=200);A.content_type=_a;return A,G
	async def handle_post_config_request(A,request):
		B=request;E,H=utils.parse_raw_path(B._message.path[A.len_prefix_running:]);F=await A._check_auth(B,E)
		if F.status==401:return F
		C,L=utils.check_http_headers(B,A.supported_media_types,accept_required=_R)
		if type(C)is web.Response:D=C;return D
		else:assert type(C)==str;G=C;assert G!=_J;M=utils.Encoding[G.rsplit(_K,1)[1]]
		try:I=await B.json()
		except json.decoder.JSONDecodeError as J:D=web.Response(status=400);K=utils.gen_rc_errors(_D,_i,error_message=_j+str(J));D.text=json.dumps(K,indent=2);return D
		return await A.handle_post_config_request_lower_half(E,H,I)
	async def handle_post_config_request_lower_half(D,data_path,query_dict,request_body):
		G=request_body;F=data_path;E=query_dict
		async with D.fifolock(Write):
			try:await D.val.handle_post_config_request(F,E,G)
			except (val.InvalidInputDocument,val.UnrecognizedQueryParameter,val.InvalidQueryParameter)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.MissingQueryParameter as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_k,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.NonexistentSchemaNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.ValidationFailed as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.ParentNodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_D,_H,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.UnrecognizedInputNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_H,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except NotImplementedError as B:A=web.Response(status=501);C=utils.gen_rc_errors(_E,_Q,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.NodeAlreadyExists as B:A=web.Response(status=409);C=utils.gen_rc_errors(_E,_l,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			if os.environ.get(_V)and _G in E:await asyncio.sleep(int(E[_G]))
			try:await D.dal.handle_post_config_request(F,E,G,D.create_callbacks,D.change_callbacks,D)
			except (dal.CreateCallbackFailed,dal.CreateOrChangeCallbackFailed,dal.ChangeCallbackFailed)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_m,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except (PluginNotFound,PluginSyntaxError,FunctionNotFound,FunctionNotCallable)as B:A=web.Response(status=501);C=utils.gen_rc_errors(_E,_Q,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except Exception as B:raise Exception(_n+str(B))
			D.val.inst=D.val.inst2;D.val.inst2=_B;await D.shared_post_commit_logic();return web.Response(status=201)
	async def handle_put_config_request(A,request):
		B=request;E,H=utils.parse_raw_path(B._message.path[A.len_prefix_running:]);F=await A._check_auth(B,E)
		if F.status==401:return F
		C,L=utils.check_http_headers(B,A.supported_media_types,accept_required=_R)
		if type(C)is web.Response:D=C;return D
		else:assert type(C)==str;G=C;assert G!=_J;M=utils.Encoding[G.rsplit(_K,1)[1]]
		try:I=await B.json()
		except json.decoder.JSONDecodeError as J:D=web.Response(status=400);K=utils.gen_rc_errors(_D,_i,error_message=_j+str(J));D.text=json.dumps(K,indent=2);return D
		return await A.handle_put_config_request_lower_half(E,H,I)
	async def handle_put_config_request_lower_half(D,data_path,query_dict,request_body):
		G=request_body;F=data_path;E=query_dict
		async with D.fifolock(Write):
			try:await D.val.handle_put_config_request(F,E,G)
			except val.InvalidDataPath as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.ParentNodeNotFound as B:A=web.Response(status=404);C=utils.gen_rc_errors(_D,_H,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.UnrecognizedInputNode as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_H,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except (val.NonexistentSchemaNode,val.ValidationFailed)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except (val.InvalidInputDocument,val.UnrecognizedQueryParameter)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_F,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.MissingQueryParameter as B:A=web.Response(status=400);C=utils.gen_rc_errors(_D,_k,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except val.NodeAlreadyExists as B:A=web.Response(status=409);C=utils.gen_rc_errors(_E,_l,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except NotImplementedError as B:A=web.Response(status=501);C=utils.gen_rc_errors(_E,_Q,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			if os.environ.get(_V)and _G in E:await asyncio.sleep(int(E[_G]))
			try:H=await D.dal.handle_put_config_request(F,E,G,D.create_callbacks,D.change_callbacks,D.delete_callbacks,D)
			except (dal.CreateCallbackFailed,dal.CreateOrChangeCallbackFailed,dal.ChangeCallbackFailed)as B:A=web.Response(status=400);C=utils.gen_rc_errors(_E,_m,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except (PluginNotFound,PluginSyntaxError,FunctionNotFound,FunctionNotCallable)as B:A=web.Response(status=501);C=utils.gen_rc_errors(_E,_Q,error_message=str(B));A.text=json.dumps(C,indent=2);return A
			except Exception as B:raise Exception("why wasn't this assertion caught by val? (assuming it's a YANG validation thing)"+str(B))
			D.val.inst=D.val.inst2;D.val.inst2=_B;await D.shared_post_commit_logic()
			if H==_P:return web.Response(status=201)
			else:return web.Response(status=204)
	async def handle_delete_config_request(A,request):
		C=request;D,G=utils.parse_raw_path(C._message.path[A.len_prefix_running:]);assert G=={};E=await A._check_auth(C,D)
		if E.status==401:return E
		B,J=utils.check_http_headers(C,A.supported_media_types,accept_required=_R)
		if type(B)is web.Response:H=B;return H
		else:
			assert type(B)==str;F=B
			if F==_J:I=_B
			else:I=utils.Encoding[F.rsplit(_K,1)[1]]
		return await A.handle_delete_config_request_lower_half(D)
	async def handle_delete_config_request_lower_half(A,data_path):
		E=data_path
		async with A.fifolock(Write):
			try:await A.val.handle_delete_config_request(E)
			except val.NonexistentSchemaNode as C:B=web.Response(status=400);D=utils.gen_rc_errors(_E,_F,error_message=str(C));B.text=json.dumps(D,indent=2);return B
			except val.NodeNotFound as C:B=web.Response(status=404);D=utils.gen_rc_errors(_D,_H,error_message=str(C));B.text=json.dumps(D,indent=2);return B
			except val.ValidationFailed as C:B=web.Response(status=400);D=utils.gen_rc_errors(_E,_F,error_message=str(C));B.text=json.dumps(D,indent=2);return B
			try:await A.dal.handle_delete_config_request(E,A.delete_callbacks,A.change_callbacks,A)
			except Exception as C:raise Exception(_n+str(C))
			A.val.inst=A.val.inst2;A.val.inst2=_B;await A.shared_post_commit_logic();return web.Response(status=204)
	async def shared_post_commit_logic(A):0
	async def handle_action_request(A,request):0
	async def handle_rpc_request(A,request):raise NotImplementedError('Native needs an RPC handler?  - client accessible!')
	def _handle_generate_symmetric_key_action(A,data_path,action_input):raise NotImplementedError(_W)
	def _handle_generate_asymmetric_key_action(A,data_path,action_input):raise NotImplementedError(_W)
	def _handle_resend_activation_email_action(A,data_path,action_input):raise NotImplementedError(_W)
	def _handle_generate_certificate_signing_request_action(A,data_path,action_input):raise NotImplementedError(_W)
async def _handle_tenant_created(watched_node_path,jsob,jsob_data_path,obj):jsob['tenant']['audit-log']={'log-entry':[]}
async def _handle_transport_changed(watched_node_path,jsob,jsob_data_path,obj):os.kill(os.getpid(),signal.SIGHUP)
async def _handle_transport_delete(watched_node_path,opaque):raise NotImplementedError('Deleting the /transport node itself cannot be constrained by YANG.')
async def _handle_plugin_created(watched_node_path,jsob,jsob_data_path,opaque):
	B=opaque;A=jsob[_U][_N];C=_o+A
	if A in B.plugins:E=sys.modules[C];del sys.modules[C];del E;del B.plugins[A]
	try:F=importlib.import_module(C)
	except ModuleNotFoundError as D:raise PluginNotFound(str(D))
	except SyntaxError as D:raise PluginSyntaxError('SyntaxError: '+str(D))
	B.plugins[A]={_p:F,_O:{}}
async def _handle_plugin_deleted(watched_node_path,opaque):C=opaque;A=re.sub(_c,_d,watched_node_path);B=_o+A;D=sys.modules[B];del sys.modules[B];del D;del C.plugins[A]
async def _handle_function_created(watched_node_path,jsob,jsob_data_path,opaque):
	B=opaque;C=re.sub(_c,_d,jsob_data_path);A=jsob[_b][_N]
	try:D=getattr(B.plugins[C][_p],A)
	except AttributeError as E:raise FunctionNotFound(str(E))
	if not callable(D):raise FunctionNotCallable("The plugin function name '"+A+"' is not callable.")
	B.plugins[C][_O][A]=D
async def _handle_function_deleted(watched_node_path,opaque):A=watched_node_path;B=opaque;C=re.sub(_c,_d,A);D=A.rsplit('=',1)[1];del B.plugins[C][_O][D]
async def _handle_admin_passwd_created(watched_node_path,jsob,jsob_data_path,obj):
	A=jsob
	def B(item):
		A=item;A[_q]=datetime.datetime.utcnow().strftime(_r)
		if _I in A and A[_I].startswith(_s):A[_I]=sha256_crypt.using(rounds=1000).hash(A[_I][3:])
	if type(A)==dict:B(A[_t])
	else:
		assert _R;assert type(A)==list
		for C in A:assert type(C)==dict;B(C)
async def _handle_admin_passwd_changed(watched_node_path,json,jsob_data_path,obj):
	def A(item):
		A=item;A[_q]=datetime.datetime.utcnow().strftime(_r)
		if _I in A and A[_I].startswith(_s):A[_I]=sha256_crypt.using(rounds=1000).hash(A[_I][3:])
		else:0
	assert json!=_B;assert jsob_data_path!=_B;A(json[_t])
async def _handle_ref_stat_parent_created(watched_node_path,jsob,jsob_data_path,obj):
	A=jsob;assert watched_node_path==jsob_data_path
	def B(item):item['reference-statistics']={'reference-count':0,'last-referenced':'never'}
	if type(A)==dict:D=next(iter(A));B(A[D])
	else:
		raise NotImplementedError('dead code?');assert type(A)==list
		for C in A:assert type(C)==dict;B(C)
def _handle_ref_stats_changed(leafrefed_node_data_path,obj):raise NotImplementedError('_handle_ref_stats_changed tested?')
def _handle_lingering_unreferenced_node_change(watched_node_path,obj):raise NotImplementedError(_u)
def _handle_expiring_certificate_change(watched_node_path,obj):raise NotImplementedError(_u)
async def _handle_asymmetric_public_key_created_or_changed(watched_node_path,jsob,jsob_data_path,obj):
	B=watched_node_path;A=jsob;I=A[_C][_e];E=base64.b64decode(I)
	if A[_C][_X]!=_v:raise dal.CreateOrChangeCallbackFailed(_w+A[_C][_X]+_Y+B.rsplit(_A,1)[0])
	try:O,J=decode_der(E,asn1Spec=rfc5280.SubjectPublicKeyInfo())
	except PyAsn1Error as D:raise dal.CreateOrChangeCallbackFailed(_x+B.rsplit(_A,1)[0]+_S+str(D)+_T)
	K=serialization.load_der_public_key(E);L=A[_C][_y];F=base64.b64decode(L)
	if A[_C][_Z]==_z:
		try:P,J=decode_der(F,asn1Spec=rfc5915.ECPrivateKey())
		except PyAsn1Error as D:raise dal.CreateOrChangeCallbackFailed('Parsing private key structure failed for '+B.rsplit(_A,1)[0]+_S+str(D)+_T)
	else:raise dal.CreateOrChangeCallbackFailed(_A0+A[_C][_Z]+_Y+B.rsplit(_A,1)[0])
	M=serialization.load_der_private_key(F,_B,_B);G=_A1;N=M.sign(G,ec.ECDSA(hashes.SHA256()))
	try:K.verify(N,G,ec.ECDSA(hashes.SHA256()))
	except InvalidSignature as D:raise dal.CreateOrChangeCallbackFailed(_A2+B.rsplit(_A,1)[0])
	if _L in A[_C]:
		if _M in A[_C][_L]:
			C=obj
			if C.dal.post_dal_callbacks is _B:C.dal.post_dal_callbacks=[]
			H=_handle_verify_asymmetric_key_and_certs_post_sweep,B.rsplit(_A,1)[0],C
			if H not in C.dal.post_dal_callbacks:C.dal.post_dal_callbacks.append(H)
async def _handle_asymmetric_private_key_created_or_changed(watched_node_path,jsob,jsob_data_path,obj):
	B=watched_node_path;A=jsob;I=A[_C][_y];E=base64.b64decode(I)
	if A[_C][_Z]==_z:O,J=decode_der(E,asn1Spec=rfc5915.ECPrivateKey())
	else:raise dal.CreateOrChangeCallbackFailed(_A0+A[_C][_Z]+_Y+B.rsplit(_A,1)[0]+_S+str(D)+_T)
	K=serialization.load_der_private_key(E,_B,_B);L=A[_C][_e];F=base64.b64decode(L)
	if A[_C][_X]!=_v:raise dal.CreateOrChangeCallbackFailed(_w+A[_C][_X]+_Y+B.rsplit(_A,1)[0])
	try:P,J=decode_der(F,asn1Spec=rfc5280.SubjectPublicKeyInfo())
	except PyAsn1Error as D:raise dal.CreateOrChangeCallbackFailed(_x+B.rsplit(_A,1)[0]+_S+str(D)+_T)
	M=serialization.load_der_public_key(F);G=_A1;N=K.sign(G,ec.ECDSA(hashes.SHA256()))
	try:M.verify(N,G,ec.ECDSA(hashes.SHA256()))
	except InvalidSignature as D:raise dal.CreateOrChangeCallbackFailed(_A2+B.rsplit(_A,1)[0])
	if _L in A[_C]:
		if _M in A[_C][_L]:
			C=obj
			if C.dal.post_dal_callbacks is _B:C.dal.post_dal_callbacks=[]
			H=_handle_verify_asymmetric_key_and_certs_post_sweep,B.rsplit(_A,1)[0],C
			if H not in C.dal.post_dal_callbacks:C.dal.post_dal_callbacks.append(H)
async def _handle_asymmetric_key_cert_created_or_changed(watched_node_path,jsob,jsob_data_path,obj):
	B=watched_node_path;I=jsob[_M][_f];J=base64.b64decode(I);K,O=decode_der(J,asn1Spec=rfc5652.ContentInfo());L=utils.degenerate_cms_obj_to_ders(K);A=[]
	for M in L:N=x509.load_der_x509_certificate(M);A.append(N)
	E=[B for B in A if B.extensions.get_extension_for_oid(ExtensionOID.BASIC_CONSTRAINTS).value.ca==_R]
	if len(E)==0:raise dal.CreateOrChangeCallbackFailed('End entity certificates must encode a certificate having "basic" constraint "ca" with value "False": '+B.rsplit(_A,1)[0])
	if len(E)>1:raise dal.CreateOrChangeCallbackFailed('End entity certificates must encode no more than one certificate having "basic" constraint "ca" with value "False" ('+str(len(E))+_A3+B.rsplit(_A,1)[0])
	G=E[0];A.remove(G);C=G
	while len(A):
		F=[B for B in A if B.subject==C.issuer]
		if len(F)==0:raise dal.CreateOrChangeCallbackFailed('End entity certificates must not encode superfluous certificates.  Found certificates unconnected to chain from the "leaf" certificate while looking for "'+str(C.subject)+_g+B.rsplit(_A,1)[0])
		if len(F)>1:raise dal.CreateOrChangeCallbackFailed('End entity certificates must not encode superfluous certificates.  CMS encodes multiple certificates having the same "subject" value ('+str(C.issuer)+'): '+B.rsplit(_A,1)[0])
		C=F[0];A.remove(C)
	D=obj
	if D.dal.post_dal_callbacks is _B:D.dal.post_dal_callbacks=[]
	H=_handle_verify_asymmetric_key_and_certs_post_sweep,B.rsplit(_A,3)[0],D
	if H not in D.dal.post_dal_callbacks:D.dal.post_dal_callbacks.append(H)
async def _handle_verify_asymmetric_key_and_certs_post_sweep(watched_node_path,conn,opaque):
	U='row_id';B=conn;A=watched_node_path;C=opaque;E=C.dal._get_row_data_for_list_path(A,B);F=re.sub('=[^/]*','',A);D=C.dal._get_jsob_for_row_id_in_table(F,E[U],B);I=D[_C][_e];J=base64.b64decode(I);K=serialization.load_der_public_key(J)
	if _L in D[_C]:
		if _M in D[_C][_L]:
			L=F+'/certificates/certificate';M=C.dal._find_rows_in_table_having_pid(L,E[U],{},B);G=M.fetchall();assert len(G)!=0
			for H in G:
				N=H['jsob'][_M][_f];O=base64.b64decode(N);P,V=decode_der(O,asn1Spec=rfc5652.ContentInfo());Q=utils.degenerate_cms_obj_to_ders(P)
				for R in Q:
					S=x509.load_der_x509_certificate(R);T=K.public_numbers()
					if S.public_key().public_numbers()==T:break
				else:raise dal.CreateOrChangeCallbackFailed('End entity certificates must encode a "leaf" certificate having a public key matching the asymmetric key\'s public key: '+A+'/certificates/certificate='+H[_N])
async def _handle_trust_anchor_cert_created_or_changed(watched_node_path,jsob,jsob_data_path,obj):
	B=watched_node_path;G=jsob[_M][_f];H=base64.b64decode(G)
	try:I,N=decode_der(H,asn1Spec=rfc5652.ContentInfo())
	except PyAsn1Error as J:raise dal.CreateOrChangeCallbackFailed('Parsing trust anchor certificate CMS structure failed for '+B.rsplit(_A,1)[0]+_S+str(J)+_T)
	K=utils.degenerate_cms_obj_to_ders(I);A=[]
	for L in K:M=x509.load_der_x509_certificate(L);A.append(M)
	D=[B for B in A if B.subject==B.issuer]
	if len(D)==0:raise dal.CreateOrChangeCallbackFailed('Trust anchor certificates must encode a root (self-signed) certificate: '+B.rsplit(_A,1)[0])
	if len(D)>1:raise dal.CreateOrChangeCallbackFailed('Trust anchor certificates must encode no more than one root (self-signed) certificate ('+str(len(D))+_A3+B.rsplit(_A,1)[0])
	F=D[0];A.remove(F);C=F
	while len(A):
		E=[B for B in A if B.issuer==C.subject]
		if len(E)==0:raise dal.CreateOrChangeCallbackFailed('Trust anchor certificates must not encode superfluous certificates.  Discovered additional certificates while looking for the issuer of "'+str(C.subject)+_g+B.rsplit(_A,1)[0])
		if len(E)>1:raise dal.CreateOrChangeCallbackFailed('Trust anchor certificates must encode a single chain of certificates.  Found '+str(len(E))+' certificates issued by "'+str(C.subject)+_g+B.rsplit(_A,1)[0])
		C=E[0];A.remove(C)
async def _check_expirations(nvh):0


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_C='/ds/ietf-datastores:operational'
_B='/ds/ietf-datastores:running'
_A=None
import os,ssl,json,base64,pyasn1,asyncio,yangson,datetime,tempfile,basicauth
from .  import utils
from aiohttp import web
from .handler import RouteHandler
from pyasn1_modules import rfc5652
from pyasn1_modules import rfc5915
from pyasn1.codec.der.decoder import decode as der_decoder
from pyasn1.codec.der.encoder import encode as der_encoder
async def set_server_header(request,response):response.headers['Server']='<redacted>'
class RestconfServer:
	root='/restconf';prefix_running=root+_B;prefix_operational=root+_C;prefix_operations=root+'/operations';len_prefix_running=len(prefix_running);len_prefix_operational=len(prefix_operational);len_prefix_operations=len(prefix_operations)
	def __init__(A,loop,dal,endpoint_config,view_handler,facade_yl=_A):
		A3='client-certs';A2='local-truststore-reference';A1='ca-certs';A0='client-authentication';z='cert-data';y='ASCII';x=':keystore/asymmetric-keys/asymmetric-key=';w='reference';v='server-identity';u='local-port';t='http';Y='tcp-server-parameters';O='certificate';N='tls-server-parameters';M='/ds/ietf-datastores:running{tail:.*}';L='/';G=dal;E='https';C=endpoint_config;B=view_handler;A.len_prefix_running=len(A.root+_B);A.len_prefix_operational=len(A.root+_C);A.loop=loop;A.dal=G;A.name=C['name'];A.view_handler=B;A.app=web.Application();A.app.on_response_prepare.append(set_server_header);A.app.router.add_get('/.well-known/host-meta',A.handle_get_host_meta);A.app.router.add_get(A.root,B.handle_get_restconf_root);A.app.router.add_get(A.root+L,B.handle_get_restconf_root);A.app.router.add_get(A.root+'/yang-library-version',B.handle_get_yang_library_version);A.app.router.add_get(A.root+'/ds/ietf-datastores:operational{tail:.*}',B.handle_get_opstate_request);A.app.router.add_get(A.root+M,B.handle_get_config_request);A.app.router.add_put(A.root+M,B.handle_put_config_request);A.app.router.add_post(A.root+M,B.handle_post_config_request);A.app.router.add_delete(A.root+M,B.handle_delete_config_request);A.app.router.add_post(A.root+'/ds/ietf-datastores:operational/{tail:.*}',B.handle_action_request);A.app.router.add_post(A.root+'/operations/{tail:.*}',B.handle_rpc_request)
		if t in C:F=t
		else:assert E in C;F=E
		A.local_address=C[F][Y]['local-address'];A.local_port=os.environ.get('SZTPD_DEFAULT_PORT',8080)
		if u in C[F][Y]:A.local_port=C[F][Y][u]
		D=_A
		if F==E:
			P=C[E][N][v][O][w]['asymmetric-key'];I=A.dal.handle_get_config_request(L+A.dal.app_ns+x+P);Z=A.loop.run_until_complete(I);Q=Z[A.dal.app_ns+':asymmetric-key'][0]['cleartext-private-key'];a=base64.b64decode(Q);b,A4=der_decoder(a,asn1Spec=rfc5915.ECPrivateKey());c=der_encoder(b);R=base64.b64encode(c).decode(y);assert Q==R;d='-----BEGIN EC PRIVATE KEY-----\n'+R+'\n-----END EC PRIVATE KEY-----\n';e=C[E][N][v][O][w][O];I=A.dal.handle_get_config_request(L+A.dal.app_ns+x+P+'/certificates/certificate='+e);f=A.loop.run_until_complete(I);g=f[A.dal.app_ns+':certificate'][0][z];h=base64.b64decode(g);i,j=der_decoder(h,asn1Spec=rfc5652.ContentInfo());k=i.getComponentByName('content');l,j=der_decoder(k,asn1Spec=rfc5652.SignedData());S=l.getComponentByName('certificates');T=''
			for m in range(len(S)):n=S[m][0];o=der_encoder(n);p=base64.b64encode(o).decode(y);T+='-----BEGIN CERTIFICATE-----\n'+p+'\n-----END CERTIFICATE-----\n'
			D=ssl.create_default_context(ssl.Purpose.CLIENT_AUTH);D.verify_mode=ssl.CERT_OPTIONAL
			with tempfile.TemporaryDirectory()as U:
				V=U+'key.pem';W=U+'certs.pem'
				with open(V,'w')as q:q.write(d)
				with open(W,'w')as r:r.write(T)
				D.load_cert_chain(W,V)
			if A0 in C[E][N]:
				H=C[E][N][A0]
				def X(truststore_ref):
					C=G.handle_get_config_request(L+G.app_ns+':truststore/certificate-bags/certificate-bag='+truststore_ref);D=A.loop.run_until_complete(C);B=[]
					for E in D[G.app_ns+':certificate-bag'][0][O]:F=base64.b64decode(E[z]);H,I=der_decoder(F,asn1Spec=rfc5652.ContentInfo());assert not I;B+=utils.degenerate_cms_obj_to_ders(H)
					return B
				J=[]
				if A1 in H:K=H[A1][A2];J+=X(K)
				if A3 in H:K=H[A3][A2];J+=X(K)
				s=utils.der_dict_to_multipart_pem({'CERTIFICATE':J});D.load_verify_locations(cadata=s)
		if F==E:assert not D is _A
		else:assert D is _A
		A.runner=web.AppRunner(A.app);A.loop.run_until_complete(A.runner.setup());A.site=web.TCPSite(A.runner,host=A.local_address,port=A.local_port,ssl_context=D,reuse_port=True);A.loop.run_until_complete(A.site.start())
	async def handle_get_host_meta(B,request):A=web.Response();A.content_type='application/xrd+xml';A.text='<XRD xmlns="http://docs.oasis-open.org/ns/xri/xrd-1.0">\n  <Link rel="restconf" href="/restconf"/>\n</XRD>';return A


# Copyright (c) 2021 Watsen Networks.  All Rights Reserved.

from __future__ import annotations
_A9='"ietf-sztp-bootstrap-server:input" is missing.'
_A8='ssl_object'
_A7='access-denied'
_A6='/ietf-sztp-bootstrap-server:report-progress'
_A5='Resource does not exist.'
_A4='Requested resource does not exist.'
_A3=':log-entry'
_A2='/devices/device='
_A1=':devices/device='
_A0='2021-01-14'
_z='2019-04-30'
_y='urn:ietf:params:xml:ns:yang:ietf-yang-types'
_x='ietf-yang-types'
_w='TBD'
_v='module'
_u='module-set-id'
_t='ietf-yang-library:modules-state'
_s='application/yang-data+xml'
_r='webhooks'
_q='/ietf-sztp-bootstrap-server:get-bootstrapping-data'
_p='Parent node does not exist.'
_o='Resource can not be modified.'
_n='2013-07-15'
_m='webhook'
_l='Unable to parse "input" document: '
_k='import'
_j='application/yang-data+json'
_i='callback'
_h='passed-input'
_g='malformed-message'
_f=':device'
_e='Content-Type'
_d='1'
_c=False
_b=':tenants/tenant='
_a='x'
_Z='implement'
_Y='reference'
_X=True
_W='invalid-value'
_V='unknown-element'
_U='application'
_T='0'
_S='path'
_R='method'
_Q='source-ip'
_P='timestamp'
_O='conformance-type'
_N='namespace'
_M='revision'
_L=':dynamic-callout'
_K='ietf-sztp-bootstrap-server:input'
_J='protocol'
_I='text/plain'
_H='name'
_G='dynamic-callout'
_F='+'
_E='return-code'
_D='error-returned'
_C=None
_B='event-details'
_A='/'
import os,json,base64,pprint,asyncio,aiohttp,yangson,datetime,basicauth,urllib.parse,pkg_resources
from .  import yl
from .  import dal
from .  import utils
from aiohttp import web
from .native import Read
from pyasn1.type import univ
from .dal import DataAccessLayer
from .rcsvr import RestconfServer
from .handler import RouteHandler
from pyasn1_modules import rfc5652
from passlib.hash import sha256_crypt
from pyasn1.codec.der.encoder import encode as encode_der
from pyasn1.codec.der.decoder import decode as der_decoder
from certvalidator import CertificateValidator,ValidationContext,PathBuildingError
from cryptography.hazmat.backends import default_backend
from cryptography import x509
class RFC8572ViewHandler(RouteHandler):
	len_prefix_running=len(RestconfServer.root+'/ds/ietf-datastores:running');len_prefix_operational=len(RestconfServer.root+'/ds/ietf-datastores:operational');len_prefix_operations=len(RestconfServer.root+'/operations');id_ct_sztpConveyedInfoXML=rfc5652._buildOid(1,2,840,113549,1,9,16,1,42);id_ct_sztpConveyedInfoJSON=rfc5652._buildOid(1,2,840,113549,1,9,16,1,43);supported_media_types=_j,_s;yl4errors={_t:{_u:_w,_v:[{_H:_x,_M:_n,_N:_y,_O:_k},{_H:'ietf-restconf',_M:'2017-01-26',_N:'urn:ietf:params:xml:ns:yang:ietf-restconf',_O:_Z},{_H:'ietf-netconf-acm',_M:'2018-02-14',_N:'urn:ietf:params:xml:ns:yang:ietf-netconf-acm',_O:_k},{_H:'ietf-sztp-bootstrap-server',_M:_z,_N:'urn:ietf:params:xml:ns:yang:ietf-sztp-bootstrap-server',_O:_Z},{_H:'ietf-yang-structure-ext',_M:'2020-06-22',_N:'urn:ietf:params:xml:ns:yang:ietf-yang-structure-ext',_O:_Z},{_H:'ietf-sztp-csr',_M:_A0,_N:'urn:ietf:params:xml:ns:yang:ietf-sztp-csr',_O:_Z},{_H:'ietf-crypto-types',_M:_A0,_N:'urn:ietf:params:xml:ns:yang:ietf-crypto-types',_O:_Z}]}};yl4conveyedinfo={_t:{_u:_w,_v:[{_H:_x,_M:_n,_N:_y,_O:_k},{_H:'ietf-inet-types',_M:_n,_N:'urn:ietf:params:xml:ns:yang:ietf-inet-types',_O:_k},{_H:'ietf-sztp-conveyed-info',_M:_z,_N:'urn:ietf:params:xml:ns:yang:ietf-sztp-conveyed-info',_O:_Z}]}}
	def __init__(A,dal,mode,yl,nvh):D='sztpd';A.dal=dal;A.mode=mode;A.nvh=nvh;B=pkg_resources.resource_filename(D,'yang');A.dm=yangson.DataModel(json.dumps(yl),[B]);A.dm4conveyedinfo=yangson.DataModel(json.dumps(A.yl4conveyedinfo),[B]);C=pkg_resources.resource_filename(D,'yang4errors');A.dm4errors=yangson.DataModel(json.dumps(A.yl4errors),[C,B])
	async def _insert_bootstrapping_log_entry(A,device_id,bootstrapping_log_entry):
		E='/bootstrapping-log';B=device_id
		if A.mode==_T:C=_A+A.dal.app_ns+':device/bootstrapping-log'
		elif A.mode==_d:C=_A+A.dal.app_ns+_A1+B[0]+E
		elif A.mode==_a:C=_A+A.dal.app_ns+_b+B[1]+_A2+B[0]+E
		D={};D[A.dal.app_ns+_A3]=bootstrapping_log_entry;await A.dal.handle_post_opstate_request(C,D)
	async def _insert_audit_log_entry(A,tenant_name,audit_log_entry):
		B=tenant_name
		if A.mode==_T or A.mode==_d or B==_C:C=_A+A.dal.app_ns+':audit-log'
		elif A.mode==_a:C=_A+A.dal.app_ns+_b+B+'/audit-log'
		D={};D[A.dal.app_ns+_A3]=audit_log_entry;await A.dal.handle_post_opstate_request(C,D)
	async def handle_get_restconf_root(D,request):
		C=request;J=_A;F=await D._check_auth(C,J)
		if type(F)is web.Response:A=F;return A
		else:H=F
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=C.remote;B[_R]=C.method;B[_S]=C.path;E,K=utils.check_http_headers(C,D.supported_media_types,accept_required=_X)
		if type(E)is web.Response:A=E;L=K;B[_E]=A.status;B[_D]=L;await D._insert_bootstrapping_log_entry(H,B);return A
		else:assert type(E)==str;G=E;assert G!=_I;I=utils.Encoding[G.rsplit(_F,1)[1]]
		A=web.Response(status=200);A.content_type=G
		if I==utils.Encoding.json:A.text='{\n    "ietf-restconf:restconf" : {\n        "data" : {},\n        "operations" : {},\n        "yang-library-version" : "2019-01-04"\n    }\n}\n'
		else:assert I==utils.Encoding.xml;A.text='<restconf xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">\n    <data/>\n    <operations/>\n    <yang-library-version>2016-06-21</yang-library-version>\n</restconf>\n'
		B[_E]=A.status;await D._insert_bootstrapping_log_entry(H,B);return A
	async def handle_get_yang_library_version(D,request):
		C=request;J=_A;F=await D._check_auth(C,J)
		if type(F)is web.Response:A=F;return A
		else:H=F
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=C.remote;B[_R]=C.method;B[_S]=C.path;E,K=utils.check_http_headers(C,D.supported_media_types,accept_required=_X)
		if type(E)is web.Response:A=E;L=K;B[_E]=A.status;B[_D]=L;await D._insert_bootstrapping_log_entry(H,B);return A
		else:assert type(E)==str;G=E;assert G!=_I;I=utils.Encoding[G.rsplit(_F,1)[1]]
		A=web.Response(status=200);A.content_type=G
		if I==utils.Encoding.json:A.text='{\n  "ietf-restconf:yang-library-version" : "2019-01-04"\n}'
		else:assert I==utils.Encoding.xml;A.text='<yang-library-version xmlns="urn:ietf:params:xml:ns:yang:ietf-restconf">2019-01-04</yang-library-version>'
		B[_E]=A.status;await D._insert_bootstrapping_log_entry(H,B);return A
	async def handle_get_opstate_request(C,request):
		D=request;E=D.path[C.len_prefix_operational:];E=_A;G=await C._check_auth(D,E)
		if type(G)is web.Response:A=G;return A
		else:I=G
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;F,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_X)
		if type(F)is web.Response:A=F;M=L;B[_E]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(I,B);return A
		else:assert type(F)==str;H=F;assert H!=_I;J=utils.Encoding[H.rsplit(_F,1)[1]]
		if E=='/ietf-yang-library:yang-library'or E==_A or E=='':A=web.Response(status=200);A.content_type=_j;A.text=getattr(yl,'sbi_rfc8572')()
		else:A=web.Response(status=404);A.content_type=H;J=utils.Encoding[A.content_type.rsplit(_F,1)[1]];K=utils.gen_rc_errors(_J,_V,error_message=_A4);N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,N);B[_D]=K
		B[_E]=A.status;await C._insert_bootstrapping_log_entry(I,B);return A
	async def handle_get_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:I=G
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_X)
		if type(E)is web.Response:A=E;M=L;B[_E]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(I,B);return A
		else:assert type(E)==str;H=E;assert H!=_I;J=utils.Encoding[H.rsplit(_F,1)[1]]
		if F==_A or F=='':A=web.Response(status=204)
		else:A=web.Response(status=404);A.content_type=H;J=utils.Encoding[A.content_type.rsplit(_F,1)[1]];K=utils.gen_rc_errors(_J,_V,error_message=_A4);N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,N);B[_D]=K
		B[_E]=A.status;await C._insert_bootstrapping_log_entry(I,B);return A
	async def handle_post_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_c)
		if type(E)is web.Response:A=E;M=L;B[_E]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_I;K=utils.Encoding[H.rsplit(_F,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_U,_W,error_message=_o)
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_J,_V,error_message=_p)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_F,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_E]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_put_config_request(C,request):
		D=request;F=D.path[C.len_prefix_running:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_c)
		if type(E)is web.Response:A=E;M=L;B[_E]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_I;K=utils.Encoding[H.rsplit(_F,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_U,_W,error_message=_o)
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_J,_V,error_message=_p)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_F,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_E]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_delete_config_request(C,request):
		D=request;G=D.path[C.len_prefix_running:];H=await C._check_auth(D,G)
		if type(H)is web.Response:A=H;return A
		else:L=H
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;E,M=utils.check_http_headers(D,C.supported_media_types,accept_required=_c)
		if type(E)is web.Response:A=E;N=M;B[_E]=A.status;B[_D]=N;await C._insert_bootstrapping_log_entry(L,B);return A
		else:
			assert type(E)==str;I=E
			if I==_I:J=_C
			else:J=utils.Encoding[I.rsplit(_F,1)[1]]
		if G==_A or G=='':A=web.Response(status=400);F=_o;K=utils.gen_rc_errors(_U,_W,error_message=F)
		else:A=web.Response(status=404);F=_p;K=utils.gen_rc_errors(_J,_V,error_message=F)
		A.content_type=I
		if J is _C:A.text=F
		else:O=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(K,J,C.dm4errors,O)
		B[_E]=A.status;B[_D]=K;await C._insert_bootstrapping_log_entry(L,B);return A
	async def handle_action_request(C,request):
		D=request;F=D.path[C.len_prefix_operational:];G=await C._check_auth(D,F)
		if type(G)is web.Response:A=G;return A
		else:J=G
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;E,L=utils.check_http_headers(D,C.supported_media_types,accept_required=_c)
		if type(E)is web.Response:A=E;M=L;B[_E]=A.status;B[_D]=M;await C._insert_bootstrapping_log_entry(J,B);return A
		else:assert type(E)==str;H=E;assert H!=_I;K=utils.Encoding[H.rsplit(_F,1)[1]]
		if F==_A or F=='':A=web.Response(status=400);I=utils.gen_rc_errors(_U,_W,error_message='Resource does not support action.')
		else:A=web.Response(status=404);I=utils.gen_rc_errors(_J,_V,error_message=_A5)
		A.content_type=H;K=utils.Encoding[A.content_type.rsplit(_F,1)[1]];N=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,K,C.dm4errors,N);B[_E]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(J,B);return A
	async def handle_rpc_request(C,request):
		P='sleep';D=request;F=D.path[C.len_prefix_operations:];J=await C._check_auth(D,F)
		if type(J)is web.Response:A=J;return A
		else:E=J
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=D.remote;B[_R]=D.method;B[_S]=D.path;H,M=utils.check_http_headers(D,C.supported_media_types,accept_required=_c)
		if type(H)is web.Response:A=H;N=M;B[_E]=A.status;B[_D]=N;await C._insert_bootstrapping_log_entry(E,B);return A
		else:
			assert type(H)==str;K=H
			if K==_I:L=_C
			else:L=utils.Encoding[K.rsplit(_F,1)[1]]
		if F==_q:
			async with C.nvh.fifolock(Read):
				if os.environ.get('SZTPD_MODE')and P in D.query:await asyncio.sleep(int(D.query[P]))
				A=await C._handle_get_bootstrapping_data_rpc(E,D,B);B[_E]=A.status;await C._insert_bootstrapping_log_entry(E,B);return A
		elif F==_A6:
			try:A=await C._handle_report_progress_rpc(E,D,B)
			except NotImplementedError as Q:raise NotImplementedError('is this ever called?')
			B[_E]=A.status;await C._insert_bootstrapping_log_entry(E,B);return A
		elif F==_A or F=='':A=web.Response(status=400);G=_A5;I=utils.gen_rc_errors(_U,_W,error_message=G)
		else:A=web.Response(status=404);G='Unrecognized RPC.';I=utils.gen_rc_errors(_J,_V,error_message=G)
		A.content_type=K
		if A.content_type==_I:A.text=G
		else:I=utils.gen_rc_errors(_J,_W,error_message=G);O=C.dm4errors.get_schema_node(_A);A.text=utils.obj_to_encoded_str(I,L,C.dm4errors,O)
		B[_E]=A.status;B[_D]=I;await C._insert_bootstrapping_log_entry(E,B);return A
	async def _check_auth(A,request,data_path):
		z='num-times-accessed';y='local-truststore-reference';x=':device-type';w='identity-certificates';v='activation-code';u='" not found for any tenant.';t='Device "';s='X-Client-Cert';e='verification';d='device-type';T='sbi-access-stats';P='lifecycle-statistics';J='comment';I='failure';H='outcome';C=request
		def F(request,supported_media_types):
			E='Accept';D=supported_media_types;C=request;B=web.Response(status=401)
			if E in C.headers and any((C.headers[E]==A for A in D)):B.content_type=C.headers[E]
			elif _e in C.headers and any((C.headers[_e]==A for A in D)):B.content_type=C.headers[_e]
			else:B.content_type=_I
			if B.content_type!=_I:F=utils.Encoding[B.content_type.rsplit(_F,1)[1]];G=utils.gen_rc_errors(_J,_A7);H=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(G,F,A.dm4errors,H)
			return B
		B={};B[_P]=datetime.datetime.utcnow();B[_Q]=C.remote;B['source-proxies']=list(C.forwarded);B['host']=C.host;B[_R]=C.method;B[_S]=C.path;K=set();M=_C;N=C.transport.get_extra_info('peercert')
		if N is not _C:O=N['subject'][-1][0][1];K.add(O)
		elif C.headers.get(s)!=_C:f=C.headers.get(s);U=bytes(urllib.parse.unquote(f),'utf-8');M=x509.load_pem_x509_certificate(U,default_backend());g=M.subject;O=g.get_attributes_for_oid(x509.ObjectIdentifier('2.5.4.5'))[0].value;K.add(O)
		Q=_C;V=_C;R=C.headers.get('AUTHORIZATION')
		if R!=_C:Q,V=basicauth.decode(R);K.add(Q)
		if len(K)==0:B[H]=I;B[J]='Device provided no identification credentials.';await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
		if len(K)!=1:B[H]=I;B[J]='Device provided mismatched authentication credentials ('+O+' != '+Q+').';await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
		E=K.pop();D=_C
		if A.mode==_T:L=_A+A.dal.app_ns+_f
		elif A.mode==_d:L=_A+A.dal.app_ns+_A1+E
		if A.mode!=_a:
			try:D=await A.dal.handle_get_opstate_request(L)
			except dal.NodeNotFound as W:B[H]=I;B[J]=t+E+u;await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
			G=_C
		else:
			try:G=await A.dal.get_tenant_name_for_global_key(_A+A.dal.app_ns+':tenants/tenant/devices/device',E)
			except dal.NodeNotFound as W:B[H]=I;B[J]=t+E+u;await A._insert_audit_log_entry(_C,B);return F(C,A.supported_media_types)
			L=_A+A.dal.app_ns+_b+G+_A2+E;D=await A.dal.handle_get_opstate_request(L)
		assert D!=_C;assert A.dal.app_ns+_f in D;D=D[A.dal.app_ns+_f]
		if A.mode!=_T:D=D[0]
		if v in D:
			if R==_C:B[H]=I;B[J]='Activation code required but none passed for serial number '+E;await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
			X=D[v];assert X.startswith('$5$')
			if not sha256_crypt.verify(V,X):B[H]=I;B[J]='Activation code mismatch for serial number '+E;await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
		else:0
		assert d in D;h=_A+A.dal.app_ns+':device-types/device-type='+D[d];Y=await A.dal.handle_get_opstate_request(h)
		if w in Y[A.dal.app_ns+x][0]:
			if N is _C and M is _C:B[H]=I;B[J]='Client cert required but none passed for serial number '+E;await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
			if N:Z=C.transport.get_extra_info(_A8);assert Z is not _C;a=Z.getpeercert(_X)
			else:assert M is not _C;a=U
			S=Y[A.dal.app_ns+x][0][w];assert e in S;assert y in S[e];b=S[e][y];i=_A+A.dal.app_ns+':truststore/certificate-bags/certificate-bag='+b['certificate-bag']+'/certificate='+b['certificate'];j=await A.dal.handle_get_config_request(i);k=j[A.dal.app_ns+':certificate'][0]['cert-data'];l=base64.b64decode(k);m,n=der_decoder(l,asn1Spec=rfc5652.ContentInfo());assert not n;o=utils.degenerate_cms_obj_to_ders(m);p=ValidationContext(trust_roots=o);q=CertificateValidator(a,validation_context=p)
			try:q._validate_path()
			except PathBuildingError as W:B[H]=I;B[J]="Client cert for serial number '"+E+"' does not validate using trust anchors specified by device-type '"+D[d]+"'";await A._insert_audit_log_entry(G,B);return F(C,A.supported_media_types)
		B[H]='success';await A._insert_audit_log_entry(G,B);r=L+'/lifecycle-statistics';c=datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
		if D[P][T][z]==0:D[P][T]['first-accessed']=c
		D[P][T]['last-accessed']=c;D[P][T][z]+=1;await A.dal.handle_put_opstate_request(r,D[P]);return E,G
	async def _handle_get_bootstrapping_data_rpc(A,device_id,request,bootstrapping_log_entry):
		Av='ietf-sztp-bootstrap-server:output';Au='ASCII';At='content';As='contentType';Ar=':configuration';Aq='configuration-handling';Ap='script';Ao='hash-value';An='hash-algorithm';Am='address';Al='referenced-definition';Ak='exited-normally';Aj='function';Ai='plugin';Ah='callout-type';Ag='serial-number';Af='rpc-supported';Ae='not';Ad='match-criteria';Ac='matched-response';Ab='input';AJ='post-configuration-script';AI='configuration';AH='pre-configuration-script';AG='os-version';AF='os-name';AE='trust-anchor';AD='port';AC='bootstrap-server';AB='ietf-sztp-conveyed-info:redirect-information';AA='data-missing';A9='response-manager';A8='operation-not-supported';z='image-verification';y='download-uri';x='boot-image';w='callback-results';v='selected-response';u='value';n=device_id;m='onboarding-information';i='error-tag';h='key';e='ietf-sztp-conveyed-info:onboarding-information';d='redirect-information';c='error';Y=request;X='ietf-restconf:errors';N='response';J='managed-response';I='response-details';E='get-bootstrapping-data-event';D='conveyed-information';C=bootstrapping_log_entry;j,AK=utils.check_http_headers(Y,A.supported_media_types,accept_required=_X)
		if type(j)is web.Response:B=j;AL=AK;C[_E]=B.status;C[_D]=AL;return B
		else:assert type(j)==str;O=j;assert O!=_I;S=utils.Encoding[O.rsplit(_F,1)[1]]
		K=_C
		if Y.body_exists:
			AM=await Y.text();AN=utils.Encoding[Y.headers[_e].rsplit(_F,1)[1]]
			try:G=A.dm.get_schema_node(_q);K=utils.encoded_str_to_obj(AM,AN,A.dm,G)
			except Exception as Z:B=web.Response(status=400);o=_l+str(Z);B.content_type=O;H=utils.gen_rc_errors(_J,_g,error_message=o);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
			if not _K in K:
				B=web.Response(status=400)
				if not _K in K:o=_l+_A9
				B.content_type=O;H=utils.gen_rc_errors(_J,_g,error_message=o);G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
		C[_B]={};C[_B][E]={};C[_B][E][_h]={}
		if K is _C:C[_B][E][_h]['no-input-passed']=[_C]
		else:
			C[_B][E][_h][Ab]=[]
			for A0 in K[_K].keys():input={};input[h]=A0;input[u]=K[_K][A0];C[_B][E][_h][Ab].append(input)
		if A.mode!=_a:P=_A+A.dal.app_ns+':'
		else:P=_A+A.dal.app_ns+_b+n[1]+_A
		if A.mode==_T:A1=P+'device'
		else:A1=P+'devices/device='+n[0]
		try:R=await A.dal.handle_get_config_request(A1)
		except Exception as Z:B=web.Response(status=501);B.content_type=_j;H=utils.gen_rc_errors(_U,A8,error_message='Unhandled exception: '+str(Z));B.text=utils.enc_rc_errors('json',H);return B
		assert R!=_C;assert A.dal.app_ns+_f in R;R=R[A.dal.app_ns+_f]
		if A.mode!=_T:R=R[0]
		if A9 not in R or Ac not in R[A9]:B=web.Response(status=404);B.content_type=O;H=utils.gen_rc_errors(_U,AA,error_message='No responses configured.');G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;C[_B][E][v]='no-responses-configured';return B
		F=_C
		for k in R[A9][Ac]:
			if not Ad in k:F=k;break
			if K is _C:continue
			for Q in k[Ad]['match']:
				if Q[h]not in K[_K]:break
				if'present'in Q:
					if Ae in Q:
						if Q[h]in K[_K]:break
					elif Q[h]not in K[_K]:break
				elif u in Q:
					if Ae in Q:
						if Q[u]==K[_K][Q[h]]:break
					elif Q[u]!=K[_K][Q[h]]:break
				else:raise NotImplementedError("Unrecognized 'match' expression.")
			else:F=k;break
		if F is _C or'none'in F[N]:
			if F is _C:C[_B][E][v]='no-match-found'
			else:C[_B][E][v]=F[_H]+" (explicit 'none')"
			B=web.Response(status=404);B.content_type=O;H=utils.gen_rc_errors(_U,AA,error_message='No matching responses configured.');G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
		C[_B][E][v]=F[_H];C[_B][E][I]={J:{}}
		if D in F[N]:
			C[_B][E][I][J]={D:{}};M={}
			if _G in F[N][D]:
				C[_B][E][I][J][D]={_G:{}};assert _Y in F[N][D][_G];p=F[N][D][_G][_Y];C[_B][E][I][J][D][_G][_H]=p;V=await A.dal.handle_get_config_request(P+'dynamic-callouts/dynamic-callout='+p);assert p==V[A.dal.app_ns+_L][0][_H];C[_B][E][I][J][D][_G][Af]=V[A.dal.app_ns+_L][0][Af];a={}
				if A.mode!=_T:a[Ag]=n[0]
				else:a[Ag]='mode-0 == no-sn'
				a['source-ip-address']=Y.remote
				if K:a['from-device']=K[_K]
				A2=Y.transport.get_extra_info(_A8)
				if A2:
					A3=A2.getpeercert(_X)
					if A3:a['identity-certificate']=A3
				if _i in V[A.dal.app_ns+_L][0]:
					C[_B][E][I][J][D][_G][Ah]=_i;A4=V[A.dal.app_ns+_L][0][_i][Ai];A5=V[A.dal.app_ns+_L][0][_i][Aj];C[_B][E][I][J][D][_G]['callback-details']={Ai:A4,Aj:A5};C[_B][E][I][J][D][_G][w]={};L=_C
					try:L=A.nvh.plugins[A4]['functions'][A5](a)
					except Exception as Z:C[_B][E][I][J][D][_G][w]['exception-thrown']=str(Z);B=web.Response(status=500);B.content_type=O;H=utils.gen_rc_errors(_U,A8,error_message='Server encountered an error while trying to generate a response: '+str(Z));G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=H;return B
					assert L and type(L)==dict
					if X in L:
						assert len(L[X][c])==1
						if any((A==L[X][c][0][i]for A in(_W,'too-big','missing-attribute','bad-attribute','unknown-attribute','bad-element',_V,'unknown-namespace',_g))):B=web.Response(status=400)
						elif any((A==L[X][c][0][i]for A in _A7)):B=web.Response(status=403)
						elif any((A==L[X][c][0][i]for A in('in-use','lock-denied','resource-denied','data-exists',AA))):B=web.Response(status=409)
						elif any((A==L[X][c][0][i]for A in('rollback-failed','operation-failed','partial-operation'))):B=web.Response(status=500)
						elif any((A==L[X][c][0][i]for A in A8)):B=web.Response(status=501)
						else:raise NotImplementedError('Unrecognized error-tag: '+L[X][c][0][i])
						B.content_type=O;H=L;G=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(H,S,A.dm4errors,G);C[_D]=L;C[_B][E][I][J][D][_G][w][Ak]='Returning an RPC-error provided by callback (NOTE: RPC-error != exception, hence a normal exit).';return B
					else:C[_B][E][I][J][D][_G][w][Ak]='Returning conveyed information provided by callback.'
				elif _r in V[A.dal.app_ns+_L][0]:C[_B][E][I][J][D][_G][Ah]=_m;raise NotImplementedError('webhooks callout support pending!')
				else:raise NotImplementedError('unhandled dynamic callout type: '+str(V[A.dal.app_ns+_L][0]))
				M=L
			elif d in F[N][D]:
				C[_B][E][I][J][D]={d:{}};M[AB]={};M[AB][AC]=[]
				if _Y in F[N][D][d]:
					f=F[N][D][d][_Y];C[_B][E][I][J][D][d]={Al:f};q=await A.dal.handle_get_config_request(P+'conveyed-information-responses/redirect-information-response='+f)
					for AO in q[A.dal.app_ns+':redirect-information-response'][0][d][AC]:
						W=await A.dal.handle_get_config_request(P+'bootstrap-servers/bootstrap-server='+AO);W=W[A.dal.app_ns+':bootstrap-server'][0];l={};l[Am]=W[Am]
						if AD in W:l[AD]=W[AD]
						if AE in W:l[AE]=W[AE]
						M[AB][AC].append(l)
				else:raise NotImplementedError('unhandled redirect-information config type: '+str(F[N][D][d]))
			elif m in F[N][D]:
				C[_B][E][I][J][D]={};M[e]={}
				if _Y in F[N][D][m]:
					f=F[N][D][m][_Y];C[_B][E][I][J][D][m]={Al:f};q=await A.dal.handle_get_config_request(P+'conveyed-information-responses/onboarding-information-response='+f);T=q[A.dal.app_ns+':onboarding-information-response'][0][m]
					if x in T:
						AP=T[x];AQ=await A.dal.handle_get_config_request(P+'boot-images/boot-image='+AP);U=AQ[A.dal.app_ns+':boot-image'][0];M[e][x]={};b=M[e][x]
						if AF in U:b[AF]=U[AF]
						if AG in U:b[AG]=U[AG]
						if y in U:
							b[y]=list()
							for AR in U[y]:b[y].append(AR)
						if z in U:
							b[z]=list()
							for A6 in U[z]:r={};r[An]=A6[An];r[Ao]=A6[Ao];b[z].append(r)
					if AH in T:AS=T[AH];AT=await A.dal.handle_get_config_request(P+'scripts/pre-configuration-script='+AS);M[e][AH]=AT[A.dal.app_ns+':pre-configuration-script'][0][Ap]
					if AI in T:AU=T[AI];A7=await A.dal.handle_get_config_request(P+'configurations/configuration='+AU);M[e][Aq]=A7[A.dal.app_ns+Ar][0][Aq];M[e][AI]=A7[A.dal.app_ns+Ar][0]['config']
					if AJ in T:AV=T[AJ];AW=await A.dal.handle_get_config_request(P+'scripts/post-configuration-script='+AV);M[e][AJ]=AW[A.dal.app_ns+':post-configuration-script'][0][Ap]
			else:raise NotImplementedError('unhandled conveyed-information type: '+str(F[N][D]))
		else:raise NotImplementedError('unhandled response type: '+str(F[N]))
		g=rfc5652.ContentInfo()
		if O==_j:g[As]=A.id_ct_sztpConveyedInfoJSON;g[At]=encode_der(json.dumps(M,indent=2),asn1Spec=univ.OctetString())
		else:assert O==_s;g[As]=A.id_ct_sztpConveyedInfoXML;G=A.dm4conveyedinfo.get_schema_node(_A);assert G;AX=utils.obj_to_encoded_str(M,S,A.dm4conveyedinfo,G);g[At]=encode_der(AX,asn1Spec=univ.OctetString())
		AY=encode_der(g,rfc5652.ContentInfo());s=base64.b64encode(AY).decode(Au);AZ=base64.b64decode(s);Aa=base64.b64encode(AZ).decode(Au);assert s==Aa;t={};t[Av]={};t[Av][D]=s;B=web.Response(status=200);B.content_type=O;G=A.dm.get_schema_node(_q);B.text=utils.obj_to_encoded_str(t,S,A.dm,G);return B
	async def _handle_report_progress_rpc(A,device_id,request,bootstrapping_log_entry):
		j='remote-port';i='wn-sztpd-rpcs:input';h='webhook-results';Y='tcp-client-parameters';X='encoding';V=device_id;U='http';L=request;H='report-progress-event';C=bootstrapping_log_entry;M,Z=utils.check_http_headers(L,A.supported_media_types,accept_required=_c)
		if type(M)is web.Response:B=M;a=Z;C[_E]=B.status;C[_D]=a;return B
		else:assert type(M)==str;I=M
		if I!=_I:R=utils.Encoding[I.rsplit(_F,1)[1]]
		if not L.body_exists:
			G='RPC "input" node missing (required for "report-progress").';B=web.Response(status=400);B.content_type=I
			if B.content_type==_I:B.text=G
			else:E=utils.gen_rc_errors(_J,_W,error_message=G);F=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(E,R,A.dm4errors,F)
			C[_D]=B.text;return B
		b=utils.Encoding[L.headers[_e].rsplit(_F,1)[1]];c=await L.text()
		try:F=A.dm.get_schema_node(_A6);N=utils.encoded_str_to_obj(c,b,A.dm,F)
		except Exception as O:B=web.Response(status=400);G=_l+str(O);B.content_type=I;E=utils.gen_rc_errors(_J,_g,error_message=G);F=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(E,R,A.dm4errors,F);C[_D]=E;return B
		if not _K in N:
			B=web.Response(status=400)
			if not _K in N:G=_l+_A9
			B.content_type=I;E=utils.gen_rc_errors(_J,_g,error_message=G);F=A.dm4errors.get_schema_node(_A);B.text=utils.obj_to_encoded_str(E,R,A.dm4errors,F);C[_D]=E;return B
		C[_B]={};C[_B][H]={};C[_B][H][_h]=N[_K];C[_B][H][_G]={}
		if A.mode==_T or A.mode==_d:J=_A+A.dal.app_ns+':preferences/notification-delivery'
		elif A.mode==_a:J=_A+A.dal.app_ns+_b+V[1]+'/preferences/notification-delivery'
		try:d=await A.dal.handle_get_config_request(J)
		except Exception as O:C[_B][H][_G]['no-callout-configured']=[_C]
		else:
			S=d[A.dal.app_ns+':notification-delivery'][_G][_Y];C[_B][H][_G][_H]=S
			if A.mode==_T or A.mode==_d:J=_A+A.dal.app_ns+':dynamic-callouts/dynamic-callout='+S
			elif A.mode==_a:J=_A+A.dal.app_ns+_b+V[1]+'/dynamic-callouts/dynamic-callout='+S
			P=await A.dal.handle_get_config_request(J);C[_B][H][_G][h]={_m:[]};T={};T[i]={};T[i]['notification']=N;e=json.dumps(T,indent=2);f='FIXME: xml output'
			if _i in P[A.dal.app_ns+_L][0]:raise NotImplementedError('callback support not implemented yet')
			elif _r in P[A.dal.app_ns+_L][0]:
				for D in P[A.dal.app_ns+_L][0][_r][_m]:
					K={};K[_H]=D[_H]
					if X not in D or D[X]=='json':W=e
					elif D[X]=='xml':W=f
					if U in D:
						Q='http://'+D[U][Y]['remote-address']
						if j in D[U][Y]:Q+=':'+str(D[U][Y][j])
						Q+='/relay-notification';K['uri']=Q
						try:
							async with aiohttp.ClientSession()as g:B=await g.post(Q,data=W)
						except aiohttp.client_exceptions.ClientConnectorError as O:K['connection-error']=str(O)
						else:
							K['http-status-code']=B.status
							if B.status==200:break
					else:assert'https'in D;raise NotImplementedError('https-based webhook is not supported yet.')
					C[_B][H][_G][h][_m].append(K)
			else:raise NotImplementedError('unrecognized callout type '+str(P[A.dal.app_ns+_L][0]))
		B=web.Response(status=204);return B